// #include <stdio.h>			// printf
// #include <string.h>			// strings in RAM
#include "../xmega/avr_compiler.h"	// Pins und Ports
#include <avr/pgmspace.h>		// strings in Flash
#include <string.h>				// memcpy
#include "../xmega/clksys_driver.h"	// softreset
#include <avr/wdt.h>			// watchdog timer

#include "control.h"			// reset, watchdog, prepareWeld, regler Menues und Dialoge
#include "global.h"				// prototypes and structur definitions
#include "eeprom.h"				// EEPROMsaveParlist()
#include "uart.h"				// UARTsendTermString()
#include "spi.h"				// Poti's, putx_MRAM()
#include "clocks.h"				// toggleTickpin()
#include "leds.h"				// LEDs_TGL()
#include "adc.h"				// Regler()
// #include "mram.h"			// MRAMSPI in setComPort


// Watchdog reset  #######################################################
/**/
void watchdog_config(void) {	// install watchdog timer, p.119
	// doesnt work:
	//CCP = CCP_IOREG_gc;	
	//WDT.CTRL = (WD_8KCLK_gc | WDT_ENABLE | WDT_CEN);
	//
	// reset after 8KCLK
	CCPWrite(&WDT.CTRL, WD_8KCLK_gc | WDT_ENABLE | WDT_CEN); // clksys_driver.h
}


// Soft Reset ############################################################
void softReset(void) {				// Software Reset p.112
	/* doesnt work:
	CCP = CCP_IOREG_gc;			// 0xd8; // config change protection p.12
	RST.CTRL |= RST_SWRST_bm;	// #define RST_SWRST_bm  0x01
	*/
	CCPWrite(&RST.CTRL, RST_SWRST_bm);	// Tip von AVRfreaks, clksys_driver.h
}


// PMIC Progable Interrupt Controller ####################################
	// #define PMIC_SET_himedlo 0b00000111
void PMIC_init(void) {
	CCP = CCP_IOREG_gc;			// 0xd8; // change protection opening p.12
	PMIC.CTRL |= PMIC_IVSEL_bm;	// open protected IOREGs for bit
	CCP = CCP_IOREG_gc;			// 0xd8; // change protection opening p.12
	PMIC.CTRL = PMIC_SET_himedlo; 	// set programmable interrupt controller (PMIC) p.128
}


void prepareRec() {			// senden rec+weld vorbereiten
	display(toggle);
	REC.ID = PRL.ID;		// ID des aktuellen Parametersatzes bernehmen
	REC.TS = 0;				// reset timestamp
	REC.SEC = 0;			// reset second
	mode = rec; 			// weld + rec
	resetTimer();			// reset both timers
	enableTimer();			// start systick and heartbeat
	// den Rest erledigen die Interrupts
}


void sendRec(void) {		// Zeitfunktionen an Lion senden
	
	// REC-struct versenden
	putx((uint8_t *) &REC, sizeof(rec_t)); // struct REC an UART senden

	REC.TS++;				// TimeStamp++ Systick-Zhler erhhen

	// live stream
	if (PRL.SEC == 0) {		// jeweils (PRL.FT Samples * 125s) bertragen
		if (REC.TS <= PRL.FT) { return;}	// weiter recorden
		else { mode = wld; }				// beenden
	}
	// nchste Sekunde
	if (REC.TS == F_SEC) { 	// ist eine Sekunde um?
		REC.TS = 0; 		// Timestamp rcksetzen
		REC.SEC++; 			// nchste Sekunde
	}
	// beenden?
	if (REC.SEC == PRL.SEC) { // Datensatz wurde vollstndig gesendet
		mode = wld;			// record beenden
	}
}


void sendParlist(void) { 	// send parameter list to PC
	genFrame();				// FR anlegen und fllen
	display(both);			// toggle display
	toggleTickpin();		// toggle systick-pin
	putx((uint8_t *) &FR, sizeof(frame_t));	// Parameterframe senden
	display(red);			// toggle display
	toggleTickpin();		// toggle systick-pin
}


void initPotis(void) {		// Register und Potis einstellen
	display(off); 			// both high
	setComPort(pot);		// auf Potis ausgeben
	SPI_POTIconfig();		// vorherige Anweisung ist nicht ausreichend! Vorsicht, unbekannter Bug
	sendPotiGains();		// Potis ber SPI einstellen
	display(red);			// red low
	setComPort(rs);			// reset to default
}


void prepareWeld(void) {	// nur Schweien
	setComPort(rs);			// reset to default
	mode = wld;				// weld
	resetTimer();			// reset both
	enableTimer();			// systick aktivieren
	sei();
}


void cmdDecode(uint8_t cmd) {	// Befehlsdekoder, 
	// wird bei Ankunft eines Befehls nur einmal durchlaufen

	switch(cmd) 
	{ 
		case d:	{					// debug: weld + record
				//PRL.SEC = FR.D;	// 
				setComPort(rs);		// debug geht nur am RS485-Port
				prepareRec();
				break;
				}	
		case w: {				// weld only						
				prepareWeld();

				break;
				}
		case l: {				// list current RAM-parameters
				sendParlist(); 	// send back to COMport
				prepareWeld();
				break;
				}
		case i: {				// install new parameters in RAM
				display(both); 
				PRL = FR.PRL;	// copy parlist from frame
				initPotis();	// Potis neu laden
				prepareWeld();	// Schweien
				break;
				}
		case b: {				// burn new parameters to EEPROM
				display(both); 
				PRL = FR.PRL;	// copy parlist from frame
				EEPROMsaveParlist(); 	// Parameterliste sichern
				softReset(); 	// reload all pars - EEPROM ist damit prfbar - Ende
				break;
				}
		case f: {				// factory reset
				genPar();		// Parametersatz generieren, incl. Potiwerte
				EEPROMsaveParlist();	// generierte Parameterliste sichern
				softReset(); 	// reload all pars - EEPROM ist damit prfbar - Ende
				//toggleTickpin(); display(red);
				//initPotis();	// Parameter auf Regs laden
				//sendParlist();	// Parliste an PC senden
				//prepareWeld(); 
				break;
				}
		case r: {				// reboot
				display(off); 
				softReset(); 
				break;
				}
		default:{				// weiterschweien
				genFrame();
				genError();
				sendParlist(); 
				resetTimer();	// reset both timers
				mode = wld;	
				enableTimer();
				break;
				}
	}
}


/*
enum com_t {	// running communication task

	rs,			// USARTC0 RS485_duplex
	sci,		// USARTE0 ABIC-SCI
	pot,		// SPIC potentiometer
	ssc,		// SPIE ABIC-SSC
	ram,		// SPID MRAM-interface
};
enum com_t com;	// globale Variable vom Typ com_t
*/



void setComPort(uint8_t comreg) {	// interface decoder

	switch(comreg) {

		case rs:{ 	// USARTC0 RS485_duplex
					REG.datareg   = &(RS_UART.DATA);
					REG.statusreg = &(RS_UART.STATUS);
					REG.rxflag = USART_RXCIF_bm;
					REG.txflag = USART_DREIF_bm;
					break;
					}
		case sci:{ 	// USARTE0 ABIC-SCI
					REG.datareg   = &(SCIUART.DATA);
					REG.statusreg = &(SCIUART.STATUS); 
					REG.rxflag = USART_RXCIF_bm;
					REG.txflag = USART_DREIF_bm;
					break;
					}
		case pot:{	// SPIC potentiometer
					REG.datareg   = &(POTSPI.DATA);
					REG.statusreg = &(POTSPI.STATUS); 
					REG.rxflag = SPI_IF_bm;
					REG.txflag = SPI_IF_bm;
					break;
					} 
		case ssc:{	// SPIE ABIC-SSC
					REG.datareg   = &(SSCSPI.DATA);
					REG.statusreg = &(SSCSPI.STATUS); 
					REG.rxflag = SPI_IF_bm;
					REG.txflag = SPI_IF_bm;
					break;
					} 
		/* case ram:{	// SPID MRAM-interface
					REG.datareg   = &(MRAMSPI.DATA);
					REG.statusreg = &(MRAMSPI.STATUS); 
					REG.rxflag = SPI_IF_bm;
					REG.txflag = SPI_IF_bm;
					break;
					} */
		default: {  // error reset
					display(off); 
					softReset(); break;
					}
	}
}


// Schweistrom schalten #####################################################

// STOP-Signal (Bypass) wird an Optokoppler geliefert
// nBY	~PC1	Optokoppler STOP ist niederohmig (Transistor leitend)

// #define BYPORT 	PORTC
// #define BYPIN 	(1<<PIN1)

void bypass_config(void) {		// Optokoppler zur Steuerung des Schweistroms

	// coupler-LED: set high (START, light off); set low (STOP, light on)
	BYPORT.DIRSET |= BYPIN; 	// direction: output, p.139
	BYPORT.OUTSET |= BYPIN;		// PC1 = 1, work
}

inline void schweissstrom(uint8_t strom) {
		// Port C, Pin 1, PC1; Quinto-X15: STOP = high, d.h. PC1 = /STOP 
	if (strom == stop) {			// ein-Flanke ist schneller als aus,
		BYPORT.OUTCLR |= BYPIN;		// Optokoppler-LED an, output: low, 
		REC.RE = MIN;				// -2048 	12 bit right adj:	anzeigen
	}
	if (strom == start)  {			// zeitlich unkritisch
		BYPORT.OUTSET |= BYPIN;		// Optokoppler-LED aus, output: high, 
		REC.RE = MAX;				// +2047 	12 bit right adj	anzeigen
	}
}



// vorgefertigte Schweiregler #################################################
// 
// Aufrufart ist Regler[1](); // ruft R1()
// Aufruf erfolgt im Detail ber den Parameter PRL.NA 
// mit Regler[PRL.NA]() in ionOS.c
// In Lion PRL.NA (Alg.-Nr.) korrekt einstellen!
// REC und PRL structs siehe global.h
//
// Spezielle Regler: siehe auch planckfunktionen.ppt
//	- Metalldampf zu Argon: Tropfenexplosion 440nm/740nm
//	- Erstarrung  Fe bei 1811 K (1538C): 1550/2300nm 
//	- Verdampfung Fe bei 3273 K (3000C): 980nm/1550nm

void R0(void) {		// Regler zur Ausgabe eines Testsignals
	static int16_t systick;	// lokaler, relativer Systick-Zhler fr Pulsdauer
	// 12 Bit ADC: Wertebereich zwischen MIN und MAX: 
	// #define MAX	0x07ff	// +2047	max 	12 Bit right adjusted
	// #define MIN	0xf800	// -2048	min 	12 Bit right adjusted

	// Es folgen 13 Ausgabekanle:
	REC.K1 = MIN;
	REC.K2 = MIN;
	REC.K3 = MIN;
	REC.K4 = MIN;
	REC.K5 = MIN;
	REC.K6 = MIN;
	REC.KU = MIN;
	REC.KI = MIN;

	REC.RA = MAX;
	REC.RB = MAX;
	REC.RC = MAX;
	REC.RD = MAX;
	REC.RE = MAX;

	schweissstrom(stop);

	systick++;		// dient der Pulsverlngerung
}

// Regler fr UV/IR-Schweissung auf K1, K2, RA, RE ############################
void R1(void) {	// Algorithmus Schweiregler
				// wird mit jedem Systick-Interrupt ausgefhrt
	static int16_t systick;	// lokaler, relativer Systick-Zhler fr Pulsdauer

	// ADC laufen im "Freerun" mit 500 kHz, 
	// alle 22s ist je ein neuer Wert da: 12 bit right-adjusted

	// ADCs lesen und Offset vorzeichenrichtig korrigieren
		REC.K1 = ADC_K1 + PRL.O1;	// 440nm	EPD440
		REC.K2 = ADC_K2 + PRL.O2;	// 950nm	LD274
		REC.K3 = ADC_K3 + PRL.O3;	// 740nm	EPD740
		REC.K4 = ADC_K4 + PRL.O4;	// 1550nm	8376
		REC.K5 = ADC_K5 + PRL.O5;	// 900nm	SFH228
		REC.K6 = ADC_K6 + PRL.O6;	// 2300nm	8423
		REC.KU = ADC_KU + PRL.OU;	// Schweispannung
		REC.KI = ADC_KI + PRL.OI;	// Schweistrom

	// 440nm/740nm: Metalldampf zu Argon: Tropfenexplosion
	// 980nm/1550nm: 	Fe Dampf
	// 1950nm/2300 nm: 	Fe Schmelze

	// Reglerwerte berechnen
		REC.RA = REC.K2 - REC.K1;		// Differenzen
		REC.RB = REC.K4 - REC.K3;		
		REC.RC = REC.K6 - REC.K5;		
		REC.RD = REC.KI - REC.KU;
	// 	REC.RE	Strom an/aus -> schweissstrom()

	// Parameterliste PRL:
		// PRL.T1:	Pulsverlngerung in Systick (8 pro ms)
		// PRL.TA:	Schwellwert fr STOP welding
		// PRL.T1:	Berechnetes STOP-Ende
		// PRL.TE: 	Schwellwert fr Stop z.B. -200 (-2048...+2047)

	// Weld-Stop prfen TE
	if (REC.RA < PRL.TE) {	// TE: Schwellwert fr Stop z.B. -200 (-2048...+2047)
		systick = 0;		// systick reset, solange stop
		schweissstrom(stop);
	};

	// Weld Restart nach Pulsverlngerung um T1 (z.B. 80 ~ 10ms)
	if(systick > PRL.T1) { 	// Pulsverlngerung in Systicks (8 pro ms)
		schweissstrom(start);	// z.B. 12 ms = 12 x 8; T1 = 96 Systicks
	}

	systick++;		// dient der Pulsverlngerung
}

// freie Regler //////////////////////////////////////////////////////

void R2(void) {;}
void R3(void) {;}
void R4(void) {;}
void R5(void) {;}
void R6(void) {;}
void R7(void) {;}
void R8(void) {;}
void R9(void) {;}


//////////////////////////////////////////////////////////////////////
